@@ -1,3 +1,8 @@
+#' @include compat-defer.R
+NULL
+
+defer_ns <- environment(defer)
+
 #' Defer Evaluation of an Expression
 #'
 #' Similar to [on.exit()], but allows one to attach
@@ -64,23 +69,7 @@
 #' defer(print("three"))
 #' deferred_clear()
 #' deferred_run()
-defer <- function(expr, envir = parent.frame(), priority = c("first", "last")) {
-  priority <- match.arg(priority)
-  if (identical(envir, .GlobalEnv) && is.null(get_handlers(envir))) {
-    message(
-      "Setting deferred event(s) on global environment.\n",
-      "  * Execute (and clear) with `deferred_run()`.\n",
-      "  * Clear (without executing) with `deferred_clear()`."
-    )
-  }
-  invisible(
-    add_handler(
-      envir,
-      handler = list(expr = substitute(expr), envir = parent.frame()),
-      front = priority == "first"
-    )
-  )
-}
+defer <- function(expr, envir = parent.frame(), priority = c("first", "last")) NULL
 
 #' @rdname defer
 #' @export
@@ -101,45 +90,11 @@ deferred_run <- function(envir = parent.frame()) {
 #' @rdname defer
 #' @export
 deferred_clear <- function(envir = parent.frame()) {
-  attr(envir, "handlers") <- NULL
+  attr(envir, "withr_handlers") <- NULL
   invisible()
 }
 
-## Handlers used for 'defer' calls. Attached as a list of expressions for the
-## 'handlers' attribute on the environment, with 'on.exit' called to ensure
-## those handlers get executed on exit.
-
-get_handlers <- function(envir) {
-  attr(envir, "handlers")
-}
-
-set_handlers <- function(envir, handlers) {
-  has_handlers <- "handlers" %in% names(attributes(envir))
-  attr(envir, "handlers") <- handlers
-  if (!has_handlers) {
-    call <- make_call(execute_handlers, envir)
-
-    # We have to use do.call here instead of eval because of the way on.exit
-    # determines its evaluation context
-    # (https://stat.ethz.ch/pipermail/r-devel/2013-November/067867.html)
-    do.call(base::on.exit, list(substitute(call), TRUE), envir = envir)
-  }
-}
-
-execute_handlers <- function(envir) {
-  handlers <- get_handlers(envir)
-  for (handler in handlers) {
-    tryCatch(eval(handler$expr, handler$envir), error = identity)
-  }
-}
-
-add_handler <- function(envir, handler, front) {
-  if (front) {
-    handlers <- c(list(handler), get_handlers(envir))
-  } else {
-    handlers <- c(get_handlers(envir), list(handler))
-  }
-
-  set_handlers(envir, handlers)
-  handler
+# Splice `compat-defer.R` into the namespace
+for (name in names(defer_ns)) {
+  assign(name, defer_ns[[name]])
 }
